/*
*   repair_ogg_vorbis.c
*/
#define VERSION "2.8.2008"
/*
*   repair ogg vorbis files by repacking of unchanged vorbis bitstream
*/
#define CPR "(c) 2008 Dr.-Ing. Ingo D. Rullhusen <d01c@uni-bremen.de>"
/*
* This program is free software; you can redistribute it and/or modify   
* it under the terms of the GNU General Public License as published by   
* the Free Software Foundation; either version 2 of the License, or   
* (at your option) any later version.   
* 
* This program is distributed in the hope that it will be useful,   
* but WITHOUT ANY WARRANTY; without even the implied warranty of   
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the   
* GNU General Public License for more details.   
*
* You should have received a copy of the GNU General Public License   
* along with this program; if not, write to the Free Software   
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
*
*/

#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <string.h>
#include <float.h>

#include <ogg/ogg.h>
#include <vorbis/codec.h>


#define OGG_SYNC_BUFF 1000


void opack_copy( ogg_packet *from, ogg_packet *to )
{
   if ( to->packet ) free( to->packet );
   *to = *from;

   to->packet = malloc( from->bytes );
   memcpy( to->packet, from->packet, from->bytes );
}


void print_help()
{
   printf("repair_ogg_vorbis (version %s)\n",VERSION);
   printf(" Copyright %s\n",CPR);
   printf(" repair_ogg_vorbis comes with ABSOLUTELY NO WARRANTY;\n");
   printf(" for details see GNU GENERAL PUBLIC LICENSE. This is free software,\n");
   printf(" and you are welcome to redistribute it under certain conditions;\n");
   printf(" see GNU GENERAL PUBLIC LICENSE for details.\n\n");
   printf("usage:\n\n");
   printf(" repair_ogg_vorbis ");
   printf("[-h] ");
   printf("[-i] ");
   printf("[-s <in stream no.>] ");
   printf("[-S <out stream no.>] ");
   printf("[-v <version integer>] ");
   printf("[-c <no. channels>] ");
   printf("[-r <sampling rate>] ");
   printf("[-U <upper bitrate hint>] ");
   printf("[-N <nominal bitrate hint>] ");
   printf("[-L <lower bitrate hint>] ");
   printf("[-W <bitrate window>] ");
   printf("<OGG in file> <OGG out file>\n");
   exit(0);
}

int main( int argc, char *argv[] )
{
   int i, j, optv, smpls;
   double bitrate;

   double bitrate_min = DBL_MAX;
   double bitrate_max = 0.0;
   int bitrate_no = 0;
   double bitrate_ava = 0.0;

   char opt_i = 0;
   char opt_s = 0; long input_stream_no = -1;
   char opt_S = 0; long output_stream_no = -1;
   char opt_v = 0; int vorbis_i_version = -1;
   char opt_c = 0; int vorbis_i_channels = -1;
   char opt_r = 0; long vorbis_i_rate = -1;
   char opt_U = 0; long vorbis_i_br_upper = -1;
   char opt_N = 0; long vorbis_i_br_nominal = -1;
   char opt_L = 0; long vorbis_i_br_lower = -1;
   char opt_W = 0; long vorbis_i_br_window = -1;

   char *in_file_name, *out_file_name;
   FILE *ogg_in, *ogg_out;

   long bytes, lrw, bytes_total, bytes_total_old, EOS, stream_no;

   oggpack_buffer ob_read, ob_write;

   ogg_sync_state ogg_y_in;
   ogg_page ogg_p_in;
   ogg_stream_state ogg_s_in;
   ogg_packet ogg_pack, opack[3];

   vorbis_dsp_state vorbis_dsp;
   vorbis_info vorbis_i_in;
   vorbis_comment vorbis_c_in;
   vorbis_block vorbis_b_in;

   ogg_sync_state ogg_y_out;
   ogg_page ogg_p_out;
   ogg_stream_state ogg_s_out;

   ogg_int64_t samples, samples_old;

   int header_no;
   char *os_buffer;


   if ( argc < 3 ) print_help();

   while ( ( optv = getopt( argc, argv, "his:S:v:c:r:U:N:L:W:" ) ) != -1 )
   {
      switch (optv)
      {
         case 'h': print_help();
         case 'i': { opt_i = 1; break; }
         case 's': { opt_s = 1; input_stream_no = atol(optarg); break; }
         case 'S': { opt_S = 1; output_stream_no = atol(optarg); break; }
         case 'v': { opt_v = 1; vorbis_i_version = atoi(optarg); break; }
         case 'c': { opt_c = 1; vorbis_i_channels = atoi(optarg); break; }
         case 'r': { opt_r = 1; vorbis_i_rate = atol(optarg); break; }
         case 'U': { opt_U = 1; vorbis_i_br_upper = atol(optarg); break; }
         case 'N': { opt_N = 1; vorbis_i_br_nominal = atol(optarg); break; }
         case 'L': { opt_L = 1; vorbis_i_br_lower = atol(optarg); break; }
         case 'W': { opt_W = 1; vorbis_i_br_window = atol(optarg); break; }
         default: print_help();
      }
   }

   in_file_name  = argv[ argc-2 ];
   out_file_name = argv[ argc-1 ];



/* init */
   ogg_in = fopen( in_file_name, "r" );
   if ( ogg_in == NULL )
   {
      fprintf( stderr, "Cannot open file \"%s\"!\n", in_file_name );
      return(1);
   }

   ogg_out = fopen( out_file_name, "w" );
   if ( ogg_out == NULL )
   {
      fprintf( stderr, "Cannot open file \"%s\"!\n", out_file_name );
      return(1);
   }

   if ( ogg_sync_init(&ogg_y_in) )
   {
      fprintf( stderr, "Error in ogg_sync_init()!\n" );
      return(1);
   }

   vorbis_info_init(&vorbis_i_in);
   vorbis_comment_init(&vorbis_c_in);

/* input stream number as default output */
   while ( ogg_sync_pageout(&ogg_y_in, &ogg_p_in) != 1 )
   {
      os_buffer = ogg_sync_buffer(&ogg_y_in, OGG_SYNC_BUFF);
      bytes = fread(os_buffer, 1, OGG_SYNC_BUFF, ogg_in);
      ogg_sync_wrote(&ogg_y_in, bytes);
   }

   stream_no = ogg_page_serialno(&ogg_p_in);

   rewind(ogg_in);
   if ( ogg_sync_clear(&ogg_y_in) )
   {   
      fprintf( stderr, "Error in ogg_sync_clear()!\n" );
      return(1);
   }


   if ( ogg_stream_init( &ogg_s_in, ( opt_s ? input_stream_no : stream_no ) ) )
   {
      fprintf( stderr, "Error in ogg_stream_init()!\n" );
      return(1);
   }


   if ( ogg_sync_init(&ogg_y_out) )
   {
      fprintf( stderr, "Error in ogg_sync_init() out!\n" );
      return(1);
   }
   if ( ogg_stream_init( &ogg_s_out, ( opt_S ? output_stream_no : stream_no ) ) )
   {
      fprintf( stderr, "Error in ogg_stream_init() out!\n" );
      return(1);
   }


/* headers */
   opack[0].packet = NULL;
   opack[1].packet = NULL;
   opack[2].packet = NULL;

/* read old */
   header_no = 0;
   while ( ! feof(ogg_in) && header_no < 3 )
   {
      while ( ogg_sync_pageout(&ogg_y_in, &ogg_p_in) != 1 )
      {
         os_buffer = ogg_sync_buffer(&ogg_y_in, OGG_SYNC_BUFF);
         bytes = fread(os_buffer, 1, OGG_SYNC_BUFF, ogg_in);
         ogg_sync_wrote(&ogg_y_in, bytes);
      }

      if( ogg_stream_pagein(&ogg_s_in, &ogg_p_in) )
      {
         fprintf( stderr, "Error in ogg_stream_pagein() (stream no.?)!\n" );
         return(1);
      }

      while ( ogg_stream_packetout(&ogg_s_in, &ogg_pack) && header_no < 3 )
      {
         if ( vorbis_synthesis_headerin(&vorbis_i_in, &vorbis_c_in, &ogg_pack) != 0 )
         {
            fprintf( stderr, "Error in vorbis_synthesis_headerin()!\n" );
            return(1);
         }

         opack_copy( &ogg_pack, &opack[header_no] );

         ++header_no;
      }
   }

   if ( header_no < 3 )
   {
      fprintf( stderr, "Error: vorbis headers not found!\n" );
      return(1);
   }

/* modify */
   samples = 0;

   if ( opt_v ) vorbis_i_in.version = vorbis_i_version;
   if ( opt_c ) vorbis_i_in.channels = vorbis_i_channels;
   if ( opt_r ) vorbis_i_in.rate = vorbis_i_rate;
   if ( opt_U ) vorbis_i_in.bitrate_upper = vorbis_i_br_upper;
   if ( opt_N ) vorbis_i_in.bitrate_nominal = vorbis_i_br_nominal;
   if ( opt_L ) vorbis_i_in.bitrate_lower = vorbis_i_br_lower;
   if ( opt_W ) vorbis_i_in.bitrate_window = vorbis_i_br_window;

   printf(" input  stream no.: %li\n",ogg_s_in.serialno);
   printf(" output stream no.: %li\n",ogg_s_out.serialno);
   printf("\n");
   printf(" VI version  %i\n",vorbis_i_in.version);
   printf(" VI channels %i\n",vorbis_i_in.channels);
   printf(" VI rate     %li\n",vorbis_i_in.rate);
   printf(" VI bitrate_upper   %li\n",vorbis_i_in.bitrate_upper);
   printf(" VI bitrate_nominal %li\n",vorbis_i_in.bitrate_nominal);
   printf(" VI bitrate_lower   %li\n",vorbis_i_in.bitrate_lower);
   printf(" VI bitrate_window  %li\n",vorbis_i_in.bitrate_window);

   printf(" VC vendor: %s\n",vorbis_c_in.vendor);
   for ( i=0; i<vorbis_c_in.comments; ++i )
   {
      printf(" VC user: %s\n",vorbis_c_in.user_comments[i]);
   }

/* write new info header */
   for ( i=0; i<3; ++i )
   {
      opack[i].granulepos = samples;
      oggpack_readinit( &ob_read, opack[i].packet, opack[i].bytes );

      if ( oggpack_read(&ob_read,8) != 0x01 ) continue;

      oggpack_writeinit(&ob_write);


      oggpack_write(&ob_write,0x01,8);

      for ( j=0; j<6; ++j )
      { 
         lrw = oggpack_read(&ob_read,8);
         oggpack_write(&ob_write,lrw,8);
      }

      lrw = oggpack_read(&ob_read,32);
      oggpack_write(&ob_write,0x00,32);
      lrw = oggpack_read(&ob_read,8);
      oggpack_write(&ob_write,vorbis_i_in.channels,8);
      lrw = oggpack_read(&ob_read,32);
      oggpack_write(&ob_write,vorbis_i_in.rate,32);   

      lrw = oggpack_read(&ob_read,32);
      oggpack_write(&ob_write,vorbis_i_in.bitrate_upper,32);
      lrw = oggpack_read(&ob_read,32);
      oggpack_write(&ob_write,vorbis_i_in.bitrate_nominal,32);
      lrw = oggpack_read(&ob_read,32);
      oggpack_write(&ob_write,vorbis_i_in.bitrate_lower,32);

      lrw = oggpack_read(&ob_read,4);
      oggpack_write(&ob_write,lrw,4);
      lrw = oggpack_read(&ob_read,4);
      oggpack_write(&ob_write,lrw,4);
      oggpack_write(&ob_write,1,1);

      if ( opack[i].bytes != oggpack_bytes(&ob_write) )
      {
         fprintf( stderr, "Error: unknown header format!\n" );
         return(1);
      }
      memcpy( opack[i].packet, ob_write.buffer, oggpack_bytes(&ob_write) );

      oggpack_writeclear(&ob_write);
      break;
   }

   ogg_stream_packetin(&ogg_s_out, &(opack[0]));
   while ( ogg_stream_pageout(&ogg_s_out, &ogg_p_out) )
   {
      bytes = fwrite(ogg_p_out.header, 1, ogg_p_out.header_len, ogg_out);
      bytes += fwrite(ogg_p_out.body, 1, ogg_p_out.body_len, ogg_out);
   }
   ogg_stream_packetin(&ogg_s_out, &(opack[1]));
   while ( ogg_stream_pageout(&ogg_s_out, &ogg_p_out) )
   {
      bytes = fwrite(ogg_p_out.header, 1, ogg_p_out.header_len, ogg_out);
      bytes += fwrite(ogg_p_out.body, 1, ogg_p_out.body_len, ogg_out);
   }
   ogg_stream_packetin(&ogg_s_out, &(opack[2]));
   while ( ogg_stream_pageout(&ogg_s_out, &ogg_p_out) )
   {
      bytes = fwrite(ogg_p_out.header, 1, ogg_p_out.header_len, ogg_out);
      bytes += fwrite(ogg_p_out.body, 1, ogg_p_out.body_len, ogg_out);
   }

   if( ogg_stream_flush(&ogg_s_out, &ogg_p_out) )
   {
      bytes = fwrite(ogg_p_out.header, 1, ogg_p_out.header_len, ogg_out);
      bytes += fwrite(ogg_p_out.body, 1, ogg_p_out.body_len, ogg_out);
   }


/* init decoder */
   if ( vorbis_synthesis_init( &vorbis_dsp, &vorbis_i_in ) )
   {
      fprintf( stderr, "Error in vorbis_synthesis_init()!\n" );
      return(1);
   }
   if ( vorbis_block_init( &vorbis_dsp, &vorbis_b_in ) )
   {
      fprintf( stderr, "Error in vorbis_block_init()!\n" );
      return(1);
   }
   if ( vorbis_synthesis_restart(&vorbis_dsp) )
   {
      fprintf( stderr, "Error in vorbis_synthesis_restart()!\n" );
      return(1);
   }


/* body */
   bytes_total = 0;
   samples_old = samples;

   printf("\n");

   while ( ! feof(ogg_in) )
   {
      while ( ogg_sync_pageout(&ogg_y_in, &ogg_p_in) != 1 && ! feof(ogg_in) )
      {
         os_buffer = ogg_sync_buffer(&ogg_y_in, OGG_SYNC_BUFF);
         bytes = fread(os_buffer, 1, OGG_SYNC_BUFF, ogg_in);
         ogg_sync_wrote(&ogg_y_in, bytes);
      }


      if( ogg_stream_pagein(&ogg_s_in, &ogg_p_in) )
      {
         fprintf( stderr, "Error in ogg_stream_pagein() (stream no.?)!\n" );
         return(1);
      }


      EOS = 0;
      while ( ogg_stream_packetout(&ogg_s_in, &ogg_pack) && EOS == 0 )
      {

         if ( vorbis_synthesis( &vorbis_b_in, &ogg_pack ) == 0 )
            vorbis_synthesis_blockin( &vorbis_dsp, &vorbis_b_in );

         while ( (smpls = vorbis_synthesis_pcmout( &vorbis_dsp, NULL )) > 0 )
         {
            samples += smpls;
            vorbis_synthesis_read( &vorbis_dsp, smpls );
         }


         ogg_pack.granulepos = samples;
         EOS = ogg_pack.e_o_s;
         ogg_stream_packetin( &ogg_s_out, &ogg_pack );

         bytes_total_old = bytes_total;
         while ( ogg_stream_pageout( &ogg_s_out, &ogg_p_out ) )
         {
            bytes = fwrite( ogg_p_out.header, 1, ogg_p_out.header_len, ogg_out );
            bytes += fwrite( ogg_p_out.body, 1, ogg_p_out.body_len, ogg_out );
            bytes_total += bytes;
         }

         if ( samples - samples_old > 0 && bytes_total - bytes_total_old > 0 )
         {
            bitrate = bytes_total * 8 * vorbis_i_in.rate
                      / (double)( samples - samples_old );

            if ( bitrate > bitrate_max ) bitrate_max = bitrate;
            if ( bitrate < bitrate_min ) bitrate_min = bitrate;
            bitrate_ava = bitrate_ava*bitrate_no + bitrate;
            ++bitrate_no;
            bitrate_ava /= (double)bitrate_no;

            samples_old = samples;
            bytes_total = 0;

            if (opt_i) printf(" pageno.: %li\t packno.: %lli\t bitrate: %f kb/s\n", ogg_page_pageno(&ogg_p_in), ogg_pack.packetno, bitrate/1000.0 );
         }

      }


      if ( ogg_sync_pageout(&ogg_y_in, &ogg_p_in) == 1 )
      {
         fprintf( stderr, "Error: OGG_SYNC_BUFF too large\n" );
         return(1);
      }
   }


/* finish */
   if ( EOS == 0 )
   {
      ogg_pack.bytes = 0;
      ogg_pack.b_o_s = 0;
      ogg_pack.e_o_s = 1;
      ogg_pack.granulepos = samples;
      ++ogg_pack.packetno;
      ogg_stream_packetin( &ogg_s_out, &ogg_pack );
   }
   bytes_total_old = bytes_total;
   if( ogg_stream_flush(&ogg_s_out, &ogg_p_out) )
   {
      bytes = fwrite(ogg_p_out.header, 1, ogg_p_out.header_len, ogg_out);
      bytes += fwrite(ogg_p_out.body, 1, ogg_p_out.body_len, ogg_out);
      bytes_total += bytes;
   }
   if ( samples - samples_old > 0 && bytes_total - bytes_total_old > 0 )
   {
      bitrate = bytes_total * 8 * vorbis_i_in.rate
                / (double)( samples - samples_old );

      if ( bitrate > bitrate_max ) bitrate_max = bitrate;
      if ( bitrate < bitrate_min ) bitrate_min = bitrate;
      bitrate_ava = bitrate_ava*bitrate_no + bitrate;
      ++bitrate_no;
      bitrate_ava /= (double)bitrate_no;

      samples_old = samples;
      bytes_total = 0;

      if (opt_i) printf(" pageno.: %li\t packno.: %lli\t bitrate: %f kb/s\n", ogg_page_pageno(&ogg_p_in), ogg_pack.packetno, bitrate/1000.0 );
   }
   if (opt_i) printf("\n");

   printf(" minimum bitrate: %f kb/s\n",bitrate_min/1000.0);
   printf(" maximum bitrate: %f kb/s\n",bitrate_max/1000.0);
   printf(" average bitrate: %f kb/s\n",bitrate_ava/1000.0);


   fclose(ogg_out);
   fclose(ogg_in);

   oggpack_writeclear(&ob_write);

   ogg_packet_clear(&opack[0]);
   ogg_packet_clear(&opack[1]);
   ogg_packet_clear(&opack[2]);

   vorbis_info_clear(&vorbis_i_in);
   vorbis_comment_clear(&vorbis_c_in);
   vorbis_block_clear(&vorbis_b_in);
   vorbis_dsp_clear(&vorbis_dsp);

   ogg_stream_clear(&ogg_s_out);
   ogg_sync_clear(&ogg_y_out);

   ogg_stream_clear(&ogg_s_in);
   ogg_sync_clear(&ogg_y_in);

   return(0);
}
